from .types import APIVersion

MAX_SUPPORTED_VERSION = APIVersion(2, 11)
#: The maximum supported protocol API version in this release

V2_MODULE_DEF_VERSION = APIVersion(2, 3)
#: The first API version in which we prefer the v2 module defs

MIN_SUPPORTED_VERSION = APIVersion(2, 0)
#: The minimum supported protocol API version in this release
